require "utilities"
require "harvesterstats"
require "harvester-anims"
require "refinery"

States = {
	Animating = 0,
	FindingOre = 1,
	MiningOre = 2,
	FindingRefinery = 3,
	ApproachedRefinery = 4,
	DroppingOre = 5,
	MovingToLocation = 6,
	FindingRefuelRefinery = 7,
	ApproachedForRefuel = 8,
	Refueling = 9,
}

StateUsesEnergy = {
	[States.Animating] = true,
	[States.FindingOre] = false,
	[States.MiningOre] = true,
	[States.FindingRefinery] = false,
	[States.ApproachedRefinery] = false,
	[States.DroppingOre] = false,
	[States.MovingToLocation] = true,
	[States.FindingRefuelRefinery] = false,
	[States.ApproachedForRefuel] = false,
	[States.Refueling] = false,
}

Harvester = {
	New = function(entity)
		local self = {
			vehicle = entity,
			animEntity = false,
			-- Movement variables.
			targetPosition = entity.position,
			targetHeading = false,
			targetDistance = 0,
			targetOrientation = 0,
			currentOrientation = entity.orientation,
			
			targetRefinery = false,
			
			state = States.FindingOre,
			oldState = false,
			onArrivalCallback = false,
			
			searchRadius = Stats.DefaultSearchRadius,
			ticksMined = 0,
			filled = false,
			refueling = false,
			
			driver = false,
			
			currentEnergy = 0,
			usingEnergy = false,
			
			animationState = false,
		}
		setmetatable(self, {__index=Harvester})
		
		-- Add a driver.
		self.driver = game.createentity{name="player", force=self.vehicle.force, position=self.vehicle.position}
		self.vehicle.passenger = self.driver
		self:SetIsFilled(false)
		
		self.animEntity = game.createentity{name = "harvester-anim", force=self.vehicle.force, position=self.vehicle.position}
		-- Get rid of the flashing "no fuel" icon.
		self.animEntity.getinventory(defines.inventory.fuel).insert({name = "coal", count = 1})
		
		self.animationState = HarvesterAnimator.New(self)
		
		self.animationState:SetOrientation(self.vehicle.orientation)
		
	
		return self
	end,
  
	Delete = function(self)
		-- Remove the driver
		if self.driver and self.driver.valid then
			self.vehicle.passenger = nil
			self.driver.destroy()
			self.driver = nil
		end
		if self.animEntity and self.animEntity.valid then
			self.animEntity.destroy()
			self.animEntity = nil
		end
		if self.vehicle and self.vehicle.valid then
			self.vehicle.destroy()
			self.vehicle = nil
		end
	end,
	
	Onload = function(self)
		setmetatable(self, {__index=Harvester})
		HarvesterAnimator.Onload(self.animationState)
	end,
		
  
	Tick = function(self)
		-- Kick any players that entered.
		self.animEntity.passenger = nil
		
		if StateUsesEnergy[self.state] then
			if not self:CheckFuel() then
				return
			end
			self.currentEnergy = self.currentEnergy - Stats.EnergyUsedPerTick
			--game.player.print("self.currentEnergy = " .. self.currentEnergy)
		end
		st = self.state
		Harvester.StateFunctions[st](self)
	end,
	
	ErrorDump = function(self)
		--game.player.print("self.currentOrientation = " .. self.currentOrientation)
		--game.player.print("self.state = " .. self.state)
		self.animationState:ErrorDump()
	end,
	
	CheckFuel = function(self)
		if self.currentEnergy <= 0 then
			self:UseFuel()
		end
		
		-- If we're running out of fuel, head for a refuel.
		if not self.refueling and not self.state == States.Animating and self.vehicle.getinventory(defines.inventory.fuel).getitemcount() < 5 then
			self:FloatingText("Heading for refuel", {g = 0.8})
			self.state = States.FindingRefuelRefinery
			self.refueling = true
		end
		
		return self.currentEnergy > 0
	end,
	
	UseFuel = function(self)
		local fuelInv = self.vehicle.getinventory(defines.inventory.fuel)
		-- We are currently low of fuel.
		if fuelInv.getitemcount() < 10 then
			-- Attempt to refuel from the hold.
			self:RefuelFromHold()
			-- There's no fuel there either.
			if fuelInv.getitemcount() < 1 then
				-- Show a notification every 2 seconds.
				if (game.tick % 120) == 0 then
					--self:FloatingText("Out of fuel", {r = 0.8})
				end
				return
			end
		end
		-- Use up some fuel.
		for fuelName, count in pairs(fuelInv.getcontents()) do
			local fuelValue = game.itemprototypes[fuelName].fuelvalue
			local fuelNeeded = math.ceil(math.max(1, -self.currentEnergy) / fuelValue)
			local fuelUsed = math.min(count, fuelNeeded)
			
			fuelInv.remove({name = fuelName, count = fuelUsed})
			self.currentEnergy = self.currentEnergy + fuelValue * fuelUsed
			break
		end
	end,
	
	RefuelFromInventory = function(self, inventory)
		local fuelName = false
		local fuelCount = 0
		-- Find out which fuel we're currently using.
		for itemName, count in pairs(self.vehicle.getinventory(defines.inventory.fuel).getcontents()) do
			if count == game.itemprototypes[itemName].stacksize then
				-- We're already full, so refueling was successful.
				return true
			end
			fuelName = itemName
			fuelCount = count
			break
		end
		-- If the fuel slot is currently empty.
		if not fuelName then
			-- Check if we have fuel in the back
			for itemName, count in pairs(inventory.getcontents()) do
				-- Is this item a fuel?
				if game.itemprototypes[itemName].fuelvalue > 0 then
					fuelName = itemName
					break
				end
			end
		end
		-- Have we found a fuel either currently used or in the back?
		if fuelName then
			-- How much is there of it in the back?
			local fuelCountInBack = inventory.getitemcount(fuelName)
			-- We need at least one, obviously.
			if fuelCountInBack > 0 then
				-- Transfer it.
				local stack = {name = fuelName, count = math.min(fuelCountInBack, game.itemprototypes[fuelName].stacksize - fuelCount)}
				self.vehicle.getinventory(defines.inventory.fuel).insert(stack)
				inventory.remove(stack)
				
				return true
			end
		end
		return false
	end,
	
	RefuelFromHold = function(self)
		self:RefuelFromInventory(self.vehicle.getinventory(2))
	end,
	
	SetIsFilled = function(self, isFilled)
		self.filled = isFilled
-- 		if self.filled then
-- 			self.driver.color = {r = 1, g = 1, b = 0, a = 0.8}
-- 		else
-- 			self.driver.color = {r = 1, g = 1, b = 1, a = 1}
-- 		end
	end,
	
	SetTargetPosition = function(self, position, onArrivalCallback)
		self.targetPosition = position
		local dPos = vector.subtract(position, self.vehicle.position)
		self.targetDistance = vector.length(dPos)
		self.targetHeading = vector.div(dPos, self.targetDistance) -- Normalized(dPos)
		if self.targetDistance > Stats.MovementSpeed then
			self.targetOrientation = DeltaposToOrientation(dPos)
		end
		self.onArrivalCallback = onArrivalCallback
		
		--game.player.print("Moving to location {" .. position.x .. ", " .. position.y .. "}")
	end,
	
	SetTargetOrientation = function(self, orientation)
		self.targetOrientation = orientation
	end,
	
	FloatingText = function(self, text, color)
		local color = color or {r = 1, g = 1, b = 1}
		game.createentity({name="flying-text", position=self.vehicle.position, text=text, color=color})
	end,
	
	FindRandomOreInRadius = function(self, radius)
		local ores = self:FindOresInRadius(radius)
		if #ores > 0 then
			return ores[math.random(#ores)]
		else
			return false
		end
	end,
	
	FindOresInRadius = function(self, radius)
		local results = game.findentitiesfiltered{type = "resource", area = GetBoundingBox(self.vehicle.position, radius)}
		local ores = {}
		for i = #results, 1, -1 do
			local ore = results[i]
			-- No fluids.
			if game.entityprototypes[ore.name].resourcecategory == "basic-fluid"
			-- No lava.
			or game.entityprototypes[ore.name].resourcecategory == "lava-magma"
			-- No depleted resources that aren't infinite.
			or (ore.amount == 0 and not game.entityprototypes[ore.name].infiniteresource)
			-- No trees.
			or string.find(ore.name, "tree") then
				table.remove(results, i)
			end
		end
		return results
	end,
	
	FindNearestUnoccupiedRefinery = function(self)
		if #glob.refineries < 1 then
			return false
		end
		local radius = 50
		local results = {}
		while #results < 1 and radius < 100 do
			results = game.findentitiesfiltered{name = "refinery-chest", area = GetBoundingBox(self.vehicle.position, radius)}
			radius = radius + 10
			for i = #results, 1, -1 do
				if Refinery.IsOccupied(results[i]) or Refinery.IsFull(results[i]) then
					table.remove(results, i)
				end
			end
		end
		local closestRefinery = FindNearestEntity(self.vehicle, results)
		return closestRefinery
	end,
	
	-- Nearly exactly the same as above, except it needn't be unoccupied and it needs to have fuel.
	FindNearestRefineryWithFuel = function(self)
		if #glob.refineries < 1 then
			return false
		end
		local radius = 50
		local results = {}
		while #results < 1 and radius < 500 do
			results = game.findentitiesfiltered{name = "refinery-chest", area = GetBoundingBox(self.vehicle.position, radius)}
			radius = radius + 10
			for i = #results, 1, -1 do
				-- Only ignore refineries without fuel if the search is less than 100.
				if radius < 100 and not Refinery.HasFuel(results[i]) then
					table.remove(results, i)
				end
			end
		end
		local closestRefinery = FindNearestEntity(self.vehicle, results)
		return closestRefinery
	end,
	
	FindNearestRefinery = function(self)
		if #glob.refineries < 1 then
			return false
		end
		local radius = 50
		local results = {}
		while #results < 1 and radius < 500 do
			results = game.findentitiesfiltered{name = "refinery-chest", area = GetBoundingBox(self.vehicle.position, radius)}
			radius = radius + 10
		end
		local closestRefinery = FindNearestEntity(self.vehicle, results)
		return closestRefinery
	end,
	
	PlayAnimation = function(self, animation)
		--game.player.print("Requesting animation (" .. animation .. ")")
		self.oldState = self.state
		self.state = States.Animating
		self.animationState:PlayAnimation(animation, function(self) self.state = self.oldState end)
	end,
	
	StateFunctions = {
		--------------------------------------------------++--------------------------------------------------
		--										  	   Animating											--
		--------------------------------------------------++--------------------------------------------------
		[States.Animating] = function(self)
			self.animationState:Tick()
		end,
		--------------------------------------------------++--------------------------------------------------
		--										  	  FindingOre											--
		--------------------------------------------------++--------------------------------------------------
		[States.FindingOre] = function(self)
			local ore = self:FindRandomOreInRadius(self.searchRadius)
			if not ore then
				self.searchRadius = self.searchRadius + 5
				return
			end
			
			-- We've found ore, so reset search radius for next time.
			self.searchRadius = Stats.DefaultSearchRadius
			
			-- Go there.
			self:SetTargetPosition(ore.position, function(self) self.state = States.MiningOre end)
			--self.nextState = States.MiningOre
			self.state = States.MovingToLocation
			
			-- We haven't mined the new location yet.
			self.scoopsMined = 0
		end,
		--------------------------------------------------++--------------------------------------------------
		--										  	   MiningOre											--
		--------------------------------------------------++--------------------------------------------------
		[States.MiningOre] = function(self)
			-- If we've mined this location enough.
			if self.scoopsMined >= Stats.ScoopsPerLocation then
				-- Find a new location.
				self.state = States.FindingOre
				self.searchRadius = Stats.CloseMineSearchRadius
				return
			end
			
			-- Mine what we can.
			local ores = self:FindOresInRadius(Stats.MiningRadius)
			local amountPerOre = math.ceil(Stats.OreMinedPerScoop / #ores)
			for _, ore in pairs(ores) do
				local oreName = ore.name
				-- Check if the ore is "special", such as silica where the resource is called "silica" and the item is called "raw-silica".
				if Stats.SpecialOres[ore.name] then
					oreName = Stats.SpecialOres[ore.name]
				end
				local oreAmount = ore.amount - math.max(0, game.entityprototypes[ore.name].minimumresourceamount)
				local maxAmount = math.min(amountPerOre, oreAmount)
				if not self.vehicle.caninsert{name = oreName, count = 1} then
					self:SetIsFilled(true)
				elseif maxAmount >= 0 then
					-- Remove up to what we can.
					if game.entityprototypes[ore.name].infiniteresource then
						-- If it's infinite we can just extract the ore we want.
						self.vehicle.insert{name = oreName, count = amountPerOre}
					elseif maxAmount > 0 then
						-- Otherwise at most that which the ore has available.
						self.vehicle.insert{name = oreName, count = maxAmount}
					end
					
					local newAmount = ore.amount - maxAmount
					
					-- If we've just depleted it and it's not infinite, destroy it.
					-- Oddly, this is not automatically done by the game.
					if newAmount == 0 then
						if not game.entityprototypes[ore.name].infiniteresource then
							ore.destroy()
						end
					else
						ore.amount = newAmount
					end
				else -- It was already negative, not for me to decide what happens...
					self.vehicle.insert{name = oreName, count = amountPerOre}
					self:FloatingText("Negative", {r = 0.8})
				end
			end
			-- If we're full, go home.
			if self.filled then
				self.scoopsMined = 0
				self.state = States.FindingRefinery
				return
			end
			self.scoopsMined = self.scoopsMined + 1
			
			self:PlayAnimation(Animations.ScoopOre)
		end,
		--------------------------------------------------++--------------------------------------------------
		--											FindingRefinery											--
		--------------------------------------------------++--------------------------------------------------
		[States.FindingRefinery] = function(self)
			local refinery = self:FindNearestUnoccupiedRefinery()
			if not refinery then
				if (game.tick % 120) == 0 then
					self:FloatingText("Cannot find unoccupied empty refinery", {r = 0.8})
				end
				return
			end
			
			self.targetRefinery = refinery
		
			-- Move towards an approach position.
			self:SetTargetPosition(vector.add(refinery.position, Stats.RefineryApproachOffset), function(self) self.state = States.ApproachedRefinery end)
			
			--self.nextState = States.ApproachedRefinery
			self.state = States.MovingToLocation
		end,
		--------------------------------------------------++--------------------------------------------------
		--										 ApproachedRefinery											--
		--------------------------------------------------++--------------------------------------------------
		[States.ApproachedRefinery] = function(self)
			-- Check if it's still there and if it's still not occupied.
			if self.targetRefinery.valid and not Refinery.IsOccupied(self.targetRefinery) then
				-- Move onto the pad.
				self:SetTargetPosition(vector.add(self.targetRefinery.position, Stats.RefineryDumpOffset), function(self) self.state = States.DroppingOre end)
				--self.nextState = States.DroppingOre
				self.state = States.MovingToLocation
			else
				-- Find another refinery.
				self.state = States.FindingRefinery
			end
		end,
		--------------------------------------------------++--------------------------------------------------
		--											  DroppingOre											--
		--------------------------------------------------++--------------------------------------------------
		[States.DroppingOre] = function(self)
			local inv = self.vehicle.getinventory(2)
			-- If the refinery is still there and we can dump the complete contents of our cargo hold.
			if self.targetRefinery.valid and Refinery.GetAvailableSlots(self.targetRefinery) > Stats.HarvesterCargoSlots then
				for itemname, count in pairs(inv.getcontents()) do
					local stack = {name = itemname, count = count}
					-- If we can dump it, dump it.
					self.targetRefinery.insert(stack)
					inv.remove(stack)
				end
			else
				--game.player.print("Cannot dump ore.")
				return
			end
			
		
			-- Wait until it's empty.
			self:SetIsFilled(false)
			self.searchRadius = Stats.DefaultSearchRadius
			
			if Refinery.HasFuel(self.targetRefinery) then
				self.state = States.Refueling
			else
				self.state = States.FindingOre
			end
			
			
			self:PlayAnimation(Animations.DumpOre)
		end,
		--------------------------------------------------++--------------------------------------------------
		--										  MovingToLocation											--
		--------------------------------------------------++--------------------------------------------------
		[States.MovingToLocation] = function(self)
			-- First rotate to face the target.
			if math.abs(self.vehicle.orientation - self.targetOrientation) > 0.001 then
				-- Check if it's faster to rotate the other way.
				if self.targetOrientation - self.vehicle.orientation > 0.5 then
					self.targetOrientation = self.targetOrientation - 1
				elseif self.targetOrientation - self.vehicle.orientation < -0.5 then
					self.targetOrientation = self.targetOrientation + 1
				end
				-- Rotate up to RotationSpeed.
				self.vehicle.orientation = self.vehicle.orientation + math.max(math.min((self.targetOrientation - self.vehicle.orientation), Stats.RotationSpeed), -Stats.RotationSpeed)
				self.animationState:SetOrientation(self.vehicle.orientation)
				return
			end
			-- We're rotated correctly, so move towards the target.
			
			dPos = vector.subtract(self.targetPosition, self.vehicle.position)
			self.targetDistance = vector.length(dPos)
			self.targetHeading = vector.div(dPos, self.targetDistance) -- vector.normalized(dPos)
			
			if self.targetDistance < Stats.MovementSpeed then
				self.vehicle.teleport(self.targetPosition)
				-- This should set the new state.
				self.onArrivalCallback(self)
				return
			end
			self.vehicle.teleport(vector.add(self.vehicle.position, vector.mul(self.targetHeading, Stats.MovementSpeed)))
			self.animEntity.teleport(self.vehicle.position)
			self.targetDistance = self.targetDistance - Stats.MovementSpeed
		end,
		--------------------------------------------------++--------------------------------------------------
		--										FindingRefuelRefinery										--
		--------------------------------------------------++--------------------------------------------------
		[States.FindingRefuelRefinery] = function(self)
			local refinery = self:FindNearestRefineryWithFuel()
			if not refinery then
				if (game.tick % 120) == 0 then
					self:FloatingText("Cannot find refinery with fuel", {r = 0.8})
				end
				return
			end
			
			self.targetRefinery = refinery
		
			-- Move towards an approach position.
			self:SetTargetPosition(vector.add(refinery.position, Stats.RefineryApproachOffset), function(self) self.state = States.ApproachedForRefuel end)
			
			--self.nextState = States.ApproachedForRefuel
			self.state = States.MovingToLocation
		end,
		--------------------------------------------------++--------------------------------------------------
		--										  ApproachedForRefuel										--
		--------------------------------------------------++--------------------------------------------------
		[States.ApproachedForRefuel] = function(self)
			-- Check if it's still there and if it's still not occupied.
			if self.targetRefinery.valid and Refinery.HasFuel(self.targetRefinery) then
				if not Refinery.IsOccupied(self.targetRefinery) then
					-- Move onto the pad.
					self:SetTargetPosition(vector.add(self.targetRefinery.position, Stats.RefineryDumpOffset), function(self) self.state = States.Refueling end)
					--self.nextState = States.Refueling
					self.state = States.MovingToLocation
				end
			else
				-- If it's invalid or has no fuel left, find another refinery.
				self.state = States.FindingRefuelRefinery
			end
		end,
		--------------------------------------------------++--------------------------------------------------
		--											  Refueling												--
		--------------------------------------------------++--------------------------------------------------
		[States.Refueling] = function(self)
			if self:RefuelFromInventory(self.targetRefinery.getinventory(defines.inventory.chest)) then
				self.refueling = false
				if self.vehicle.getinventory(2).getitemcount() > 0 then
					self.state = States.DroppingOre
				else
					self.state = States.FindingOre
				end
				
				return
			end
			-- We didn't find any fuel, so find another refinery.
			self.state = States.FindingRefuelRefinery
		end,
	},
}
