require "utilities"

local beltAreas = {
	N = { -- North.
		{-1, -4}, {2, -3}
	},
	--E = { -- East.
	--	{3, -3}, {4, 3}
	--},
	--S = { -- South.
	--	--{-3, 3}, {3, 4}
	--},
	--W = { -- West.
	--	{-3, -3}, {-4, 3}
	--},
}

local beltDirections = {
	N = 0,
	E = 2,
	S = 4,
	W = 8,
}

local laneOff = 0.25
local dropOff = 0.35
local dropOffsets = {
	N = {{-laneOff,  dropOff}, { laneOff,  dropOff}},
	E = {{-dropOff, -laneOff}, {-dropOff,  laneOff}},
	S = {{ laneOff, -dropOff}, {-laneOff, -dropOff}},
	W = {{ dropOff,  laneOff}, { dropOff, -laneOff}},
}

Refinery = {
	--------------------------------------------------++--------------------------------------------------
	--										   Static functions											--
	--------------------------------------------------++--------------------------------------------------
	-- Static functions.
	IsOccupied = function(refineryEntity)
		local results = game.findentitiesfiltered{name = "harvester", area = GetBoundingBox(vector.add(refineryEntity.position, Stats.RefineryDumpOffset), 1)}
		return #results > 0
	end,

	HasFuel = function(refineryEntity)
		for itemName, count in pairs(refineryEntity.getinventory(defines.inventory.chest).getcontents()) do
			-- Is this item a fuel?
			if game.itemprototypes[itemName].fuelvalue > 0 then
				return true
			end
		end
		return false
	end,

	GetAvailableSlots = function(refineryEntity)
		if Refinery.IsFull(refineryEntity) then return 0 end
		local inv = refineryEntity.getinventory(defines.inventory.chest)

		local totalSlots = inv.getbar()
		local slotsTaken = GetOccupiedSlots(inv)
		return math.max(0, totalSlots - slotsTaken)
	end,

	-- Returns whether or not there is a free slot in the refinery.
	IsFull = function(refineryEntity)
		return not refineryEntity.getinventory(defines.inventory.chest).caninsert({name = "refinery", count = 1})
	end,


	--------------------------------------------------++--------------------------------------------------
	--										   Class functions											--
	--------------------------------------------------++--------------------------------------------------
	New = function(entity)
		local self = {
			refinery = entity,
			chest = false,
			inventory = false,

			-- Make sure not all refineries check for belts in the same tick.
			tickOffset = game.tick % 20,

			hasBelts = false,
			belts = {N = {}, E = {}, S = {}, W = {}},
		}
		setmetatable(self, {__index=Refinery})
		
		-- It's placed by placing "refinery", so place the "refinery-chest".
		self.chest = game.createentity({name="refinery-chest", force=self.refinery.force, position=self.refinery.position})
		self.inventory = self.chest.getinventory(defines.inventory.chest)

		return self
	end,

	Delete = function(self)
		if self.refinery and self.refinery.valid then
			self.refinery.destroy()
		end
		if self.chest and self.chest.valid then
			self.chest.destroy()
		end
	end,
	
	Onload = function(self)
		setmetatable(self, {__index=Refinery})
	end,

	Tick = function(self)
		if ((game.tick + self.tickOffset) % 20) == 0 then
			self:CheckForBelts()
		end

		if self.hasBelts and self.inventory.getitemcount() > 0 then
			self:DropOnBelts()
		end
	end,

	CheckForBelts = function(self)
		self.hasBelts = false
		self.belts = {N = {}, E = {}, S = {}, W = {}}
		for dir, area in pairs(beltAreas) do
			local results = game.findentitiesfiltered{type = "transport-belt", area = {vector.add(self.refinery.position, area[1]), vector.add(self.refinery.position, area[2])}}
			for _, belt in pairs(results) do
				if belt.direction == beltDirections[dir] then
					table.insert(self.belts[dir], belt)
					self.hasBelts = true
				end
			end
		end
	end,

	DropOnBelts = function(self)
		for dir, belts in pairs(self.belts) do
			for _, belt in pairs(belts) do
				if not belt.valid then
					self:CheckForBelts()
					return
				end
				self:DropOnBelt(belt, dir)
			end
		end
	end,

	DropOnBelt = function(self, belt, dir)
		local dropped = 0
		for itemName, itemCount in pairs(self.inventory.getcontents()) do
			-- If it contains fuel.
			local count = itemCount
			if game.itemprototypes[itemName].fuelvalue > 0 then
				-- Keep at least a stack.
				count = count - game.itemprototypes[itemName].stacksize
			end
			local stack = {name = itemName, count = 1}

			-- If there's two of the current item, drop it on both lanes if possible.
			if count > 1 then
				dropped = 2
				for _, offset in pairs(dropOffsets[dir]) do
					local possiblePos = game.findnoncollidingposition("item-on-ground", vector.add(belt.position, offset), 0.01, 0.01)
					if possiblePos then
						game.createentity{position = possiblePos, name = "item-on-ground", stack=stack}
						self.inventory.remove(stack)
					end
				end
			-- Otherwise drop one. The next item loop will drop on the other side.
			elseif count == 1 then
				dropped = dropped + 1
				local offset = dropOffsets[dir][dropped]
				local possiblePos = game.findnoncollidingposition("item-on-ground", vector.add(belt.position, offset), 0.01, 0.01)
				if possiblePos then
					game.createentity{position = possiblePos, name = "item-on-ground", stack=stack}
					self.inventory.remove(stack)
				end
			end

			if dropped == 2 then
				break
			end
		end
	end,
}
